
# THIS FILE IS GENERATED FROM linkahead SETUP.PY
#
short_version = '0.15.1'
version = '0.15.1'
full_version = '0.15.1'
git_revision = '86ffb96add53e78c0de8f5024a072d974f958adb'
release = True

if not release:
    version = full_version
